import socket
import json
import logging
from colorlog import ColoredFormatter
from Crypto.Cipher import DES, Blowfish, ARC4, PKCS1_OAEP
from Crypto.PublicKey import RSA
from Crypto.Util.Padding import pad
from Crypto.Random import get_random_bytes
import base64

# Add USER INPUT logging to system
USER_INPUT = 25
logging.addLevelName(USER_INPUT, "USER INPUT")

def user_input(msg, *args, **kwargs):
    logging.log(USER_INPUT, msg, *args, **kwargs)

# Configure colored logging
log_formatter = ColoredFormatter(
    "%(asctime)s - %(log_color)s%(levelname)-12s%(reset)s - %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
    log_colors={
        'DEBUG': 'cyan',
        'INFO': 'green',
        'WARNING': 'yellow',
        'ERROR': 'red',
        'CRITICAL': 'magenta',
        'USER INPUT': 'blue',
    }
)

# Show logs on terminal
console_handler = logging.StreamHandler()
console_handler.setFormatter(log_formatter)

logging.basicConfig(level=logging.DEBUG, handlers=[console_handler])

# Send data to server
def send_message_to_server(host, port, algorithm, key, key_length, encrypted_message, encrypted_key):
    data = {
        'algorithm': algorithm,
        'key_length': key_length,
        'encrypted_key': encrypted_key,
        'encrypted_message': encrypted_message
    }
    
    # Start connection to server and send data
    logging.info("Connecting to server...")
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        try:
            s.connect((host, port))
            logging.info(f"Connected to server at {host}:{port}")
            s.sendall(json.dumps(data).encode())
            logging.info("Data sent to server.")
        except Exception as e:
            logging.error(f"Connection error: {e}")

# Encrypt user message according to selected algorithm and key length
def encrypt_message(algorithm, key, message):
    logging.info(f"Starting encryption with algorithm: {algorithm}")
    
    if algorithm == 'DES':
        cipher = DES.new(key, DES.MODE_CBC)
        ciphertext = cipher.encrypt(pad(message.encode(), DES.block_size))
        encrypted_message = base64.b64encode(cipher.iv + ciphertext).decode()
        logging.info("DES encryption successful.")
        return encrypted_message
    
    elif algorithm == 'Blowfish':
        cipher = Blowfish.new(key, Blowfish.MODE_CBC)
        ciphertext = cipher.encrypt(pad(message.encode(), Blowfish.block_size))
        encrypted_message = base64.b64encode(cipher.iv + ciphertext).decode()
        logging.info("Blowfish encryption successful.")
        return encrypted_message
    
    elif algorithm == 'RC4':
        cipher = ARC4.new(key)
        ciphertext = cipher.encrypt(message.encode())
        encrypted_message = base64.b64encode(ciphertext).decode()
        logging.info("RC4 encryption successful.")
        return encrypted_message
    
    else:
        logging.error("Unsupported encryption algorithm.")
        raise ValueError("Unsupported encryption algorithm")

# Encrypt the key using RSA public key
def encrypt_key_with_rsa(public_key, key):
    cipher = PKCS1_OAEP.new(public_key)
    encrypted_key = cipher.encrypt(key)
    return base64.b64encode(encrypted_key).decode()

# Load the RSA public key (for encrypting the symmetric key)
def load_public_key():
    with open('../keys/public.pem', 'rb') as public_file:
        public_key = RSA.import_key(public_file.read())
    return public_key

def main():
    host = '192.168.178.63'
    port = 12345
    
    # Load the public RSA key (for encrypting the symmetric key)
    public_key = load_public_key()
    
    while True:
        # User input for encryption settings
        logging.info("Starting the client...")
        
        print("Choose encryption algorithm:")
        print("1. DES")
        print("2. Blowfish")
        print("3. RC4")
        
        encryption_choice = int(input("Select: "))
        
        if encryption_choice == 1:
            algorithm = 'DES'
            key = get_random_bytes(8)
            key_length = 64
        
        elif encryption_choice == 2:
            algorithm = 'Blowfish'
            print("Choose key length:")
            print("1. 64-bit")
            print("2. 128-bit")
            key_length_choice = int(input("Select: "))
            
            if key_length_choice == 1:
                key = get_random_bytes(8)
                key_length = 64
            elif key_length_choice == 2:
                key = get_random_bytes(16)
                key_length = 128
            else:
                logging.error("Invalid choice for key length.")
                print("Invalid choice!")
                continue
            
        elif encryption_choice == 3:
            algorithm = 'RC4'
            print("Choose key length:")
            print("1. 64-bit")
            print("2. 128-bit")
            key_length_choice = int(input("Select: "))
            
            if key_length_choice == 1:
                key = get_random_bytes(8)
                key_length = 64
            elif key_length_choice == 2:
                key = get_random_bytes(16)
                key_length = 128
            else:
                logging.error("Invalid choice for key length.")
                print("Invalid choice!")
                continue
        
        else:
            logging.error("Invalid choice for encryption algorithm.")
            print("Invalid choice!")
            continue
        
        # Get user message
        message = input("Message: ")
        
        logging.info(f"Selected: {algorithm}, Key length: {key_length}")
        encrypted_message = encrypt_message(algorithm, key, message)
        
        # Encrypt the key with the RSA public key
        encrypted_key = encrypt_key_with_rsa(public_key, key)
        
        # Send the data to the server
        send_message_to_server(host, port, algorithm, key, key_length, encrypted_message, encrypted_key)
        
        # Connection open or closed
        again = input("Do you want to send another message? (y/n): ").strip().lower()
        if again != 'y':
            logging.info("Exiting client...")
            break

if __name__ == '__main__':
    main()