import socket
import json
import logging
from colorlog import ColoredFormatter
from Crypto.Cipher import DES, Blowfish, ARC4, PKCS1_OAEP
from Crypto.PublicKey import RSA
from Crypto.Util.Padding import unpad
import base64

# Add USER INPUT logging to system
USER_INPUT = 25
logging.addLevelName(USER_INPUT, "USER INPUT")

def user_input(msg, *args, **kwargs):
    logging.log(USER_INPUT, msg, *args, **kwargs)

# Configure colored logging
log_formatter = ColoredFormatter(
    "%(asctime)s - %(log_color)s%(levelname)-12s%(reset)s - %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
    log_colors={
        'DEBUG': 'cyan',
        'INFO': 'green',
        'WARNING': 'yellow',
        'ERROR': 'red',
        'CRITICAL': 'magenta',
        'USER INPUT': 'blue',
    }
)

# Show logs on terminal
console_handler = logging.StreamHandler()
console_handler.setFormatter(log_formatter)

logging.basicConfig(level=logging.DEBUG, handlers=[console_handler])

# Load RSA private key (for decrypting the symmetric key)
def load_private_key():
    with open('../keys/private.pem', 'rb') as private_file:
        private_key = RSA.import_key(private_file.read())
    return private_key

# Load RSA public key (for encrypting the symmetric key)
def load_public_key():
    with open('../keys/public.pem', 'rb') as public_file:
        public_key = RSA.import_key(public_file.read())
    return public_key

# Manage client connection
def handle_client(conn, addr, private_key):
    logging.info(f"Handling client {addr}")
    
    # Receive up to 1024 Bytes data from client
    data = conn.recv(1024).decode()
    logging.info(f"Received data from client {addr}.")
    
    # Manage received JSON data and extract information
    try:
        data = json.loads(data)
        algorithm = data['algorithm']
        key_length = data['key_length']
        encrypted_key = base64.b64decode(data['encrypted_key'])
        encrypted_message = data['encrypted_message']
        logging.info(f"Algorithm: {algorithm}, Key Length: {key_length}")
        user_input(f"Received data from client {addr}: Algorithm: {algorithm}, Key Length: {key_length}, Encrypted message: {encrypted_message}")
    except json.JSONDecodeError as e:
        logging.error(f"Error decoding data: {e}")
        conn.close()
        return
    
    # Decrypt the symmetric key using RSA
    key = decrypt_symmetric_key(private_key, encrypted_key)
    logging.info(f"Decrypted symmetric key from client {addr}.")
    
    # Decrypt the message using the symmetric key
    decrypted_message = decrypt_message(algorithm, key, encrypted_message)
    logging.info(f"Decrypted message from client {addr}: {decrypted_message}")
    
    # Respond to the client
    response = {
        "status": "success",
        "decrypted_message": decrypted_message
    }
    conn.sendall(json.dumps(response).encode())
    logging.info(f"Sent response to client {addr}")

# Decrypt the symmetric key using RSA private key
def decrypt_symmetric_key(private_key, encrypted_key):
    cipher = PKCS1_OAEP.new(private_key)
    key = cipher.decrypt(encrypted_key)
    logging.info(f"Decrypted symmetric key: {key.hex()}")
    return key

# Decrypt received message using the symmetric key
def decrypt_message(algorithm, key, encrypted_message):
    logging.info(f"Starting decryption with algorithm: {algorithm}")
    
    # Transform received message into binaries
    encrypted_message = base64.b64decode(encrypted_message)
    logging.info(f"Encrypted message: {encrypted_message.hex()}")
    
    if algorithm == 'DES':
        iv = encrypted_message[:8]
        ciphertext = encrypted_message[8:]
        logging.info(f"IV: {iv.hex()}, Ciphertext: {ciphertext.hex()}")
        cipher = DES.new(key, DES.MODE_CBC, iv)
        decrypted_message = unpad(cipher.decrypt(ciphertext), DES.block_size).decode()
        logging.info("DES decryption successful.")
        return decrypted_message
    
    elif algorithm == 'Blowfish':
        iv = encrypted_message[:8]
        ciphertext = encrypted_message[8:]
        cipher = Blowfish.new(key, Blowfish.MODE_CBC, iv)
        decrypted_message = unpad(cipher.decrypt(ciphertext), Blowfish.block_size).decode()
        logging.info("Blowfish decryption successful.")
        return decrypted_message
    
    elif algorithm == 'RC4':
        cipher = ARC4.new(key)
        decrypted_message = cipher.decrypt(encrypted_message).decode()
        logging.info("RC4 decryption successful.")
        return decrypted_message
    
    else:
        logging.error("Unsupported algorithm.")
        raise ValueError("Unsupported decryption algorithm")

# Define server configuration
def main():
    host = '192.168.178.63'
    port = 12345
    
    # Load RSA private key
    private_key = load_private_key()
    
    # Wait for incoming connections
    logging.info("Starting server...")
    
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        try:
            s.bind((host, port))
            s.listen()
            logging.info(f"Listening on {host}:{port}...")
        except Exception as e:
            logging.error(f"Server error: {e}")
            return
        
        # Accept and close connection to client
        while True:
            try:
                conn, addr = s.accept()
                logging.info(f"Connection from {addr}")
                handle_client(conn, addr, private_key)
                conn.close()
            except Exception as e:
                logging.error(f"Error handling connection: {e}")
                continue

if __name__ == '__main__':
    main()
